package nom.tam.util;

/*
 * #%L
 * nom.tam FITS library
 * %%
 * Copyright (C) 1996 - 2024 nom-tam-fits
 * %%
 * This is free and unencumbered software released into the public domain.
 *
 * Anyone is free to copy, modify, publish, use, compile, sell, or
 * distribute this software, either in source code form or as a compiled
 * binary, for any purpose, commercial or non-commercial, and by any
 * means.
 *
 * In jurisdictions that recognize copyright laws, the author or authors
 * of this software dedicate any and all copyright interest in the
 * software to the public domain. We make this dedication for the benefit
 * of the public at large and to the detriment of our heirs and
 * successors. We intend this dedication to be an overt act of
 * relinquishment in perpetuity of all present and future rights to this
 * software under copyright law.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * #L%
 */

import java.util.logging.Logger;

/**
 * Access to {@link Logger} instances by class. Many of the FITS classes log minor issues they encounter during
 * processing (such as small FITS standard violations in 3rd pary FITS files). This tool provides access to these class
 * logs simply by the class itself. E.g. to get the log of {@link nom.tam.fits.Header}, you'd simply call:
 * 
 * <pre>
 * Logger headerLog = LoggerHelper(Header.class);
 * </pre>
 * 
 * without needing to know what name the log is being stored under. You can then use this to access the log, or to mute
 * unwanted messages on a per-class basis.
 */
public final class LoggerHelper {

    private LoggerHelper() {
        // Not needed
    }

    /**
     * @param  aClass the class to get the logger for.
     *
     * @return        a Java {@code Logger} for the specified class.
     */
    public static Logger getLogger(Class<?> aClass) {
        return Logger.getLogger(aClass.getName());
    }
}
