package nom.tam.util;

/*-
 * #%L
 * nom.tam FITS library
 * %%
 * Copyright (C) 1996 - 2024 nom-tam-fits
 * %%
 * This is free and unencumbered software released into the public domain.
 *
 * Anyone is free to copy, modify, publish, use, compile, sell, or
 * distribute this software, either in source code form or as a compiled
 * binary, for any purpose, commercial or non-commercial, and by any
 * means.
 *
 * In jurisdictions that recognize copyright laws, the author or authors
 * of this software dedicate any and all copyright interest in the
 * software to the public domain. We make this dedication for the benefit
 * of the public at large and to the detriment of our heirs and
 * successors. We intend this dedication to be an overt act of
 * relinquishment in perpetuity of all present and future rights to this
 * software under copyright law.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * #L%
 */

import java.io.IOException;

import org.junit.Test;

import nom.tam.fits.FitsFactory;

public class ArrayDataOutputTest {

    @Test
    public void testHDUWriteNonFitsStream() throws Exception {
        int[][] data = new int[2][2];

        ArrayDataOutput o = new ArrayDataOutput() {

            @Override
            public void write(int b) throws IOException {
            }

            @Override
            public void write(byte[] b) throws IOException {
            }

            @Override
            public void write(byte[] b, int off, int len) throws IOException {
            }

            @Override
            public void writeBoolean(boolean v) throws IOException {
            }

            @Override
            public void writeByte(int v) throws IOException {
            }

            @Override
            public void writeBytes(String s) throws IOException {
            }

            @Override
            public void writeChar(int v) throws IOException {
            }

            @Override
            public void writeChars(String s) throws IOException {
            }

            @Override
            public void writeDouble(double v) throws IOException {
            }

            @Override
            public void writeFloat(float v) throws IOException {
            }

            @Override
            public void writeInt(int v) throws IOException {
            }

            @Override
            public void writeLong(long v) throws IOException {
            }

            @Override
            public void writeShort(int v) throws IOException {
            }

            @Override
            public void writeUTF(String s) throws IOException {
            }

            @Override
            public void close() throws IOException {
            }

            @Override
            public void flush() throws IOException {
            }

            @Override
            public void write(boolean[] buf, int offset, int size) throws IOException {
            }

            @Override
            public void write(char[] buf, int offset, int size) throws IOException {
            }

            @Override
            public void write(double[] buf, int offset, int size) throws IOException {
            }

            @Override
            public void write(float[] buf, int offset, int size) throws IOException {
            }

            @Override
            public void write(int[] buf, int offset, int size) throws IOException {
            }

            @Override
            public void write(long[] buf, int offset, int size) throws IOException {
            }

            @Override
            public void write(short[] buf, int offset, int size) throws IOException {
            }

            @Override
            public void write(String[] buf, int offset, int size) throws IOException {
            }

            @Override
            public void writeArray(Object o) throws IOException, IllegalArgumentException {
            }

        };

        FitsFactory.hduFactory(data).write(o);

        // All is good if wse got this far without an exception.
    }
}
