package nom.tam.fits.compression.provider.param.rice;

/*
 * #%L
 * nom.tam FITS library
 * %%
 * Copyright (C) 1996 - 2024 nom-tam-fits
 * %%
 * This is free and unencumbered software released into the public domain.
 *
 * Anyone is free to copy, modify, publish, use, compile, sell, or
 * distribute this software, either in source code form or as a compiled
 * binary, for any purpose, commercial or non-commercial, and by any
 * means.
 *
 * In jurisdictions that recognize copyright laws, the author or authors
 * of this software dedicate any and all copyright interest in the
 * software to the public domain. We make this dedication for the benefit
 * of the public at large and to the detriment of our heirs and
 * successors. We intend this dedication to be an overt act of
 * relinquishment in perpetuity of all present and future rights to this
 * software under copyright law.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * #L%
 */

import nom.tam.fits.compression.algorithm.api.ICompressOption;
import nom.tam.fits.compression.algorithm.rice.RiceCompressOption;
import nom.tam.fits.compression.provider.param.api.ICompressHeaderParameter;
import nom.tam.fits.compression.provider.param.base.CompressParameters;

/**
 * (<i>for internal use</i>) Rice compression parameters that are recorded in a FITS compressed HDU.
 */
@SuppressWarnings({"javadoc", "deprecation"})
public class RiceCompressParameters extends CompressParameters {

    private RiceBlockSizeParameter blockSize;

    private RiceBytePixParameter bytePix;

    public RiceCompressParameters(RiceCompressOption option) {
        blockSize = new RiceBlockSizeParameter(option);
        bytePix = new RiceBytePixParameter(option);
    }

    @Override
    public RiceCompressParameters copy(ICompressOption option) {
        if (option instanceof RiceCompressOption) {
            RiceCompressOption ro = (RiceCompressOption) option;
            RiceCompressParameters p = (RiceCompressParameters) super.clone();

            p.blockSize = (RiceBlockSizeParameter) blockSize.copy(ro);
            p.bytePix = (RiceBytePixParameter) bytePix.copy(ro);

            return p;
        }
        return null;
    }

    @Override
    protected ICompressHeaderParameter[] headerParameters() {
        return new ICompressHeaderParameter[] {blockSize, bytePix};
    }
}
