package nom.tam.fits.compression.provider.param.api;

import nom.tam.fits.Header;
import nom.tam.fits.HeaderCardException;

/*
 * #%L
 * nom.tam FITS library
 * %%
 * Copyright (C) 1996 - 2024 nom-tam-fits
 * %%
 * This is free and unencumbered software released into the public domain.
 *
 * Anyone is free to copy, modify, publish, use, compile, sell, or
 * distribute this software, either in source code form or as a compiled
 * binary, for any purpose, commercial or non-commercial, and by any
 * means.
 *
 * In jurisdictions that recognize copyright laws, the author or authors
 * of this software dedicate any and all copyright interest in the
 * software to the public domain. We make this dedication for the benefit
 * of the public at large and to the detriment of our heirs and
 * successors. We intend this dedication to be an overt act of
 * relinquishment in perpetuity of all present and future rights to this
 * software under copyright law.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * #L%
 */

/**
 * (<i>for internal use</i>) Compression parameter that must be stored along the header meta data of the hdu.
 */
public interface ICompressHeaderParameter extends ICompressParameter {

    /**
     * get the value from the header and set it in the compression option.
     * 
     * @param      header the header of the hdu
     * 
     * @deprecated        Use {@link #getValueFromHeader(Header)} instead.
     */
    default void getValueFromHeader(IHeaderAccess header) {
        getValueFromHeader(header.getHeader());
    }

    /**
     * Get the parameter value from the option and set it into the fits header.
     * 
     * @param      header the header to add the parameter.
     * 
     * @deprecated        Use {@link #setValueInHeader(Header)} instead
     */
    default void setValueInHeader(IHeaderAccess header) {
        setValueInHeader(header.getHeader());
    }

    /**
     * get the value from the header and set it in the compression option.
     * 
     * @param  header              the header of the hdu
     * 
     * @throws HeaderCardException if there was a problem accessing the header
     */
    void getValueFromHeader(Header header) throws HeaderCardException;

    /**
     * Get the parameter value from the option and set it into the fits header.
     * 
     * @param  header              the header to add the parameter.
     * 
     * @throws HeaderCardException if there was a problem accessing the header
     */
    void setValueInHeader(Header header) throws HeaderCardException;
}
